/*
 *  Copyright (C) 2005  Thadeu Lima de Souza Cascardo <cascardo@holoscopio.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */



#ifndef GRAMMAR_H
#define GRAMMAR_H

#include <glib.h>

typedef gint (*nextcb) (gpointer, gpointer*);

typedef struct
{
  gboolean terminal;
  GQuark value;
} symbol_t;

typedef struct _rule rule_t;
typedef struct
{
  GHashTable* grammar;
} grammar_t;

symbol_t* symbol_new (gboolean, GQuark);
symbol_t* symbol_copy (symbol_t*);
guint symbol_hash (gconstpointer);
gboolean symbol_equal (gconstpointer, gconstpointer);
gint symbol_cmp (symbol_t*, symbol_t*);

rule_t* rule_new ();
rule_t* rule_copy (rule_t*);
gint rule_cmp (rule_t*, rule_t*);
gboolean rule_equal (gconstpointer, gconstpointer);
guint rule_hash (gconstpointer);
symbol_t* rule_pop (rule_t*);
void rule_append (rule_t*, symbol_t*);
void rule_delete (rule_t*);

grammar_t* grammar_new ();
rule_t* grammar_rule_new (grammar_t*, symbol_t*);
GList* grammar_get_rules (grammar_t*, symbol_t*);
GList* grammar_get_rule (rule_t*);
void grammar_delete (grammar_t*);

#endif
