/*
 *  Copyright (C) 2005  Thadeu Lima de Souza Cascardo <cascardo@holoscopio.com>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 */



#include <rdp.h>
#include <stdlib.h>

struct _buffer
{
  symbol_t* symbol;
  gpointer attrib;
};

gpointer leaf_new (gpointer data)
{
  return g_node_new (data);
}

gpointer tree_new (rule_t* rule)
{
  return g_node_new (rule);
}

gpointer tree_add (gpointer tree, gpointer data)
{
  return g_node_append (tree, data);
}

void tree_delete (gpointer tree)
{
  g_node_destroy (tree);
}

rdp_t* rdp_new (nextcb cb, gpointer data, gint value, grammar_t* grammar)
{

  rdp_t* parser;

  parser = g_malloc (sizeof (rdp_t));

  parser->cb = cb;
  parser->data = data;
  parser->start = symbol_new (FALSE, value);
  parser->grammar = grammar;

  parser->buffer = g_list_append (NULL, NULL);

  return parser;

}

void rdp_delete (rdp_t* rdp)
{
  g_free (rdp->start);
}

symbol_t* buffer_next (rdp_t* parser, gpointer* attrib)
{

  buffer_t* buffer;

  if (parser->buffer->next == NULL)
    {
      buffer = g_malloc (sizeof (buffer_t));
      buffer->symbol = g_malloc (sizeof (symbol_t));
      buffer->symbol->terminal = TRUE;
      buffer->symbol->value = parser->cb (parser->data, &(buffer->attrib));
      g_list_append (parser->buffer, buffer);
    }

  parser->buffer = g_list_next (parser->buffer);
  buffer = (buffer_t*) parser->buffer->data;

  if (attrib)
    *attrib = buffer->attrib;

  return buffer->symbol;
      
}

gboolean rdp_step (rdp_t* parser, symbol_t* symbol, gpointer* attrib)
{

  GList* l;
  GList* buffer;
  gpointer attr;

  buffer = parser->buffer;

  if (symbol != NULL && symbol->terminal)
    {
      symbol_t* s;
      s = buffer_next (parser, &attr);
      if (!symbol_equal (symbol, s))
	{
	  parser->buffer = buffer;
	  return FALSE;
	}
      *attrib = leaf_new (attr);
      return TRUE;
    }

  l = grammar_get_rules (parser->grammar, symbol);
  for (; l != NULL; l = g_list_next (l))
    {

      rule_t* rule;
      GList* m;

      rule = (rule_t*) l->data;


      *attrib = tree_new (symbol_copy (symbol));

      m = grammar_get_rule (rule);

      /*
      if (m == NULL)
	{
	  tree_add (*attrib, leaf_new (g_string_new ("")));
	  return TRUE;
	}
      */

      while (m != NULL)
	{

	  symbol_t* s;

	  s = (symbol_t*) m->data;

	  if (!rdp_step (parser, s, &attr))
	    {
	      parser->buffer = buffer;
	      break;
	    }

	  tree_add (*attrib, attr);

	  m = g_list_next (m);

	}

      if (m == NULL)
	return TRUE;
      else
	tree_delete (*attrib);

    }

  return FALSE;

}

gpointer rdp_build (rdp_t* parser)
{

  gpointer attrib;

  if (rdp_step (parser, parser->start, &attrib))
    return attrib;

  return NULL;

}
